/*
#       lidsadm.c --- The Linux Intrusion Detection System Configuration Tool 
#       (C) Huagang Xie 1999-2001 All rights reserved.
#       EMail complaints to xie@gnuchina.org
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
*/

/* ------------------------------------------------------------------------- */

/* Includes */

#include <linux/config.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/capability.h>
#include <linux/lidsext.h>
#include <linux/kdev_t.h>
#include <string.h>
#include "lids_cap.h"

/* ------------------------------------------------------------------------- */

/* Defines */

/* LIDS commands */
#define LIDS_NONE         -1
#define LIDS_ADD	   1
#define LIDS_DELETE	   2
#define LIDS_LIST	   3
#define LIDS_DELETE_ALL	   4
#define LIDS_MK_PASSWD	   5
#define LIDS_SWITCH	   6
#define LIDS_INIT	   7
#define LIDS_UPDATE	   8
#define LIDS_VIEW	   9

/* LIDS access types */
#define LIDS_DENY	   0
#define LIDS_READ	   1
#define LIDS_APPEND 	   3
#define LIDS_WRITE         7
#define LIDS_IGNORE 	   7
#define LIDS_EXTEND        8
#define LIDS_CAP 	   16

#define LIDS_DELETE_FLAG   100

#define LIDS_CONF_FILE	   "/etc/lids/lids.conf"
#define LIDS_PW_FILE	   "/etc/lids/lids.pw"
#define LIDS_LOCKS         "/proc/sys/lids/locks"

#define LIDS_CONF_COMMENT  "#\n# \tThis file is auto generated by lidsadm \n#\tPlease do not modify this file by hand\n#\n"

#define LIDS_MAGIC_1       0x004e6748
#define LIDS_MAGIC_2       0x68002d68
#define LIDS_MAGIC_3       0xe68400c6
#define LIDS_MAGIC_4       0xd94aac00

#define CAP_FULL_SET            (~0)
#define CAP_TO_MASK(flag)       (1 << (flag))
#define cap_raise(cap, flag)    (cap |= CAP_TO_MASK(flag))
#define cap_lower(cap, flag)    (cap &= ~CAP_TO_MASK(flag))
#define cap_raised(cap, flag)   (cap & CAP_TO_MASK(flag) & CAP_FULL_SET)
#define flag_raise(flags, flag) (flags |= CAP_TO_MASK(flag))
#define flag_lower(flags, flag) (flags &= ~CAP_TO_MASK(flag))

/* ------------------------------------------------------------------------- */

/* Typedefs and globals */

char lids_type_desc[6][7] = {"DENY","READ","APPEND","WRITE","IGNORE","GRANT\0"};
int  lids_type_val[6] = {LIDS_DENY,LIDS_READ,LIDS_APPEND,LIDS_WRITE,LIDS_WRITE,LIDS_CAP};
char lids_special_target_val[2] = {0,1};

typedef __u32 kernel_cap_t;

typedef char passwd_t[64];

typedef struct lids_locks_s {
	int magic1;
	kernel_cap_t cap_bset;
	int magic2;
	lids_flags_t flags;
	int magic3;
	passwd_t passwd;
	int magic4;
} lids_locks_t;

static char *theflags[] =
{
  "LIDS_GLOBAL",  "de-/activate LIDS entirely",
  "",             "",
  "RELOAD_CONF",  "reload config. file and inode/dev of protected programs",
  "LIDS",         "de-/activate LIDS locally (the shell & childs)",
  NULL,           NULL,
};

struct lids{
	int	           type;	  /* access control type */
	unsigned long int  s_ino;	  /* subject inode number */
	unsigned int	   s_dev;	  /* subject dev number */
	char 	           s_file[1024];  /* subject filename */
	unsigned long int  o_ino;	  /* object inode number */
	unsigned int	   o_dev;	  /* object dev number*/
	char 	           o_file[1024];  /* oject filename */
	int	           inherit;	  /* inherit level */
};

struct lids lids[1024];

char    *program;
int	last;

/* ------------------------------------------------------------------------- */

/* Prototypes */

void exit_error (int status, char *msg);
void exit_version();
void exit_normal();
void exit_help();
void lids_add_file(char *subject_file,char *object_file,int type,int inherit,int);
void lids_del_file(char *,char*);
void lids_del_all();
void lids_list_file(int type);
void lids_make_rmd160_passwd();
int  lids_verify();
void lids_switch(int optind, int argc, char *argv[]);
void lids_init(int optind, int argc, char *argv[]);
void lids_update();
#ifdef VIEW
void lids_view();
#endif

/* ------------------------------------------------------------------------- */

main(int argc,char **argv)
{
	int	command = LIDS_NONE;
	int 	type=LIDS_NONE;
	char 	subject_file[1024],object_file[1024];
	char 	del_file[1024];
	char 	sw_flag[16];
	int	c,i;
	int	cap_type=0; /* program capability type */
	int	inherit=0;/* if the acl is inheritable*/
	int	domain = 0;/* here define the DOMAIN */

	memset(subject_file,'\0',1024);
	memset(object_file,'\0',1024);
	program=argv[0];
	if(getuid()!=0) {
		exit_error(2, "you must be root to run this program");
	}
#ifdef VIEW
	while ((c = getopt (argc, argv, "VUADZLSIPz:s:o:j:i:dehv")) != -1){
#else
	while ((c = getopt (argc, argv, "UADZLSIPz:s:o:j:i:dehv")) != -1){
#endif
        	switch (c)
			{
#ifdef VIEW
			case 'V':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_VIEW;
				break;
#endif
		        case 'U':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_UPDATE;
				break;
			case 'A':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_ADD;
				break;
			case 'D':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_DELETE;
				break;
			case 'L':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_LIST;
				break;
			case 'e':
				if (command != LIDS_LIST)
					exit_error (2, "error commands specified");
				type = LIDS_EXTEND;
				break;
			case 'Z':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_DELETE_ALL;
				break;
			case 'S':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_SWITCH;
				break;
			case 'I':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_INIT;
				break;
			case 'o':	/* for object */
				if (command == LIDS_NONE)
					exit_error (2, "no comands specified");
				strcpy(object_file,optarg);
				if(strchr(object_file,':'))
					exit_error (2, "filename can not contain :");
				break;
			case 's':	/* program capability subject*/
				if (command == LIDS_NONE)
					exit_error (2, "no comands specified");
				strcpy(subject_file,optarg);
				if(strchr(object_file,':'))
					exit_error (2, "filename must not contain :");
				break;
			case 'i':
				if (command == LIDS_NONE)
					exit_error (2, "no comands specified");
				inherit= atoi(optarg);
				break;
			case 'd':
				if(command == LIDS_NONE)
					exit_error (2, "no comands specified");
				domain=1;
				break;
			case 'j':
				if (command == LIDS_NONE)
					exit_error (2, "no comands specified");

				if ( type != LIDS_NONE )
					exit_error (2, "multiple types specified");
				type = -1;
				
				for(i=0;i<6;i++) {
					if( !strcmp(optarg,lids_type_desc[i])) {
						type=lids_type_val[i];
					}
				}
				if(type == -1)
					exit_error(2,"cap type must be READ, WRITE, APPEND, DENY or GRANT");
				break;
			case 'P':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_MK_PASSWD;
				break;
			case 'v':
				exit_version();
				break;
			case 'h':
				exit_help();
			default:
			}
	}
	
	if ((command != LIDS_INIT) && (command != LIDS_SWITCH) && (optind < argc))
        	exit_error (2, "unknown arguments found on commandline");
	if ( command == LIDS_NONE || argc < 2 )
		exit_normal();
	if ( command == LIDS_ADD & type == LIDS_NONE) 
		exit_normal();
	if ( subject_file[0] != 0  && object_file[0] == 0 && command == LIDS_ADD) 
		exit_error(2,"defined subject must be related to an object");

	switch(command) {
	case LIDS_ADD:
		printf("ADD\n");
		
		lids_add_file(subject_file,object_file,type,inherit,domain);
		
		break;
	case LIDS_DELETE:
		printf("DELETE\n");
		lids_del_file(subject_file,object_file);
		break;
	case LIDS_LIST:
		printf("LIST\n");
		lids_list_file(type);
		break;
	case LIDS_DELETE_ALL:
		printf("DELETE_ALL\n");
		lids_del_all();
		break;
	case LIDS_MK_PASSWD:
		printf("MAKE PASSWD\n");
		lids_make_rmd160_passwd();
		break;
	case LIDS_SWITCH:
		printf("SWITCH\n");
		lids_switch(optind,argc,argv);
		break;
	case LIDS_INIT:
		printf("INIT\n");
		lids_init(optind,argc,argv);
		break;
	case LIDS_UPDATE:
		printf("UPDATE\n");
		lids_update();
		break;
#ifdef VIEW
	case LIDS_VIEW:
		printf("VIEW\n");
		lids_view();
		break;
#endif
	}
	exit(0);
}

void lids_del_all(char *filename,int type)
{
	FILE *fd;
	
	if((fd = fopen(LIDS_CONF_FILE,"w"))== NULL ) {
		exit_error(3,"cannot open "LIDS_CONF_FILE);
	}
	fwrite(LIDS_CONF_COMMENT,1,strlen(LIDS_CONF_COMMENT),fd);
	fclose(fd);
}

/*
 *	lids_get_str get the ino,dev,filename from the buff.
 *	the format of buffer is "inode:dev:filename"
 */
int lids_get_str(char *buffer,unsigned long * ino,int *dev,char *filename)
{
	char    *p,*q;
	char 	ino_str[1024];
	char 	ino_dev[1024];

	memset(ino_str,'\0',1024);
	memset(ino_dev,'\0',1024);

	p = q = NULL;
	p = strchr(buffer,':');	
	if ( p == NULL ) return -1; 
	strncpy(ino_str,buffer,p-buffer);
	p++;
	q = strchr(p+1,':');
	if ( q == NULL ) return -1; 
	strncpy(ino_dev,p,q-p);

	strcpy(filename,q+1);
	*ino = atol(ino_str);
	*dev = atoi(ino_dev);
	return 0;
}
/*
 * check_format check if the given buf is the valid string for current version
 */
int check_format(char *buf)
{
	int i,j;
	char *p=NULL;

	i=0;
	p = strchr(buf,':');
	while(p!=NULL) {
		p = strchr(p+1,':');
		i++;
	}
	if(i!=7) return -1;
	return 0;
}
/*
 *  lids read conf read /etc/lids.conf into a data structure.
 */ 
void lids_read_conf()
{
	FILE 	*fd;
	int	i=0,type;
	char 	buffer[1024];
	char    *p,*q;
		
	if((fd = fopen(LIDS_CONF_FILE,"r"))== NULL ) {
		exit_error(3,"cannot open "LIDS_CONF_FILE);
	}
	while(fgets(buffer,1024,fd) && i < 1024) {
		if ( buffer[0] == '#' ) continue;
		if ( buffer[strlen(buffer)-1] == '\n' )
			buffer[strlen(buffer)-1] = '\0' ;
		if(check_format(buffer) < 0) {
			printf("syntax error in %s\n", LIDS_CONF_FILE);
			exit(-1);
		}
		p = strchr(buffer,':');	
		if ( p == NULL ) continue; 
		q = strchr(p+1,':');
		if ( q == NULL ) continue; 
		p = strchr(q+1,':');	
		if ( p == NULL || *(p+1) == ':' ) continue; 
		*p = '\0';
		/* get the permission type */
		q = strchr(p+1,':');
		if(q==NULL) continue;
		
		*q='\0';
		lids[i].type = atoi(p+1);

		/* get the inherit level */ 	
		p = strchr(q+1,':');
		if ( p == NULL || *(p+1) == ':' ) continue;
		*p = '\0';
		lids[i].inherit = atoi(q+1);
		if (lids_get_str(buffer,&lids[i].s_ino,&lids[i].s_dev,lids[i].s_file) <0 ) continue;
		
		if (lids_get_str(p+1,&lids[i].o_ino,&lids[i].o_dev,lids[i].o_file) <0 ) continue;
#if 0	
		printf("inode= %d, dev= %d, file= %s, type= %d, inherit= %d, o_inode= %ld, o_dev= %d, o_file= %s\n",lids[i].s_ino,
                                                                                                                    lids[i].s_dev,
                                                                                                                    lids[i].s_file,
                                                                                                                    lids[i].type,
                                                                                                                    lids[i].inherit,
                                                                                                                    lids[i].o_ino,
                                                                                                                    lids[i].o_dev,
                                                                                                                    lids[i].o_file);
#endif
		i++;
	}
	last = i;

	fclose(fd);
}

void lids_write_file()
{
	int i;
	FILE *fd;
	
	if((fd = fopen(LIDS_CONF_FILE,"w"))== NULL ) {
		exit_error(3,"cannot open "LIDS_CONF_FILE);
	}
	fwrite(LIDS_CONF_COMMENT,1,strlen(LIDS_CONF_COMMENT),fd);
	for(i=0;i<last;i++) {
		if(lids[i].type != LIDS_DELETE_FLAG )
			fprintf(fd,"%d:%d:%s:%d:%d:%d:%d:%s\n",lids[i].s_ino,
                                                               lids[i].s_dev,
                                                               lids[i].s_file,
                                                               lids[i].type,
                                                               lids[i].inherit,
                                                               lids[i].o_ino,
                                                               lids[i].o_dev,
                                                               lids[i].o_file);
	}

	fclose(fd);
}
/*
 *	get the str desc number from type number
 */ 
int get_type_str(int type)
{
	int k,j;
	
	k = -1;
	for(j=0;j<6;j++) {
		if(type == lids_type_val[j]) {
			k=j;
			break;
		}
	}
	if(k == -1 ) exit_error(2,"type mismatch in "LIDS_CONF_FILE);
	return k;
}

void lids_list_file(int type)
{
	int i,j,k;
	char s_file[1024],o_file[1024];
	
	memset(s_file,'\0',1024);
	memset(o_file,'\0',1024);
	lids_read_conf();
#ifdef DEBUG
	printf("last = %d \n",last);
#endif
	if ( type == LIDS_EXTEND) 
		printf("    Inode   \t       Dev     ");
	printf("\tSubject \t ACCESS(inherit) \t Object\n");
	printf("     -----------------------------------------------------\n");
	for(i=0;i<last;i++) {
		if(*lids[i].s_file == '\0' )
			strcpy(s_file,"Any File");
		else
			strcpy(s_file,lids[i].s_file);
		if(*lids[i].o_file == '\0' )
			strcpy(o_file,"Any File");
		else
			strcpy(o_file,lids[i].o_file);
			
		if(lids[i].type < 0 ){
			type = get_type_str(-lids[i].type);
			printf("\t%10s\t%8s(domain):%d\t%10s\n",s_file,lids_type_desc[type],lids[i].inherit,o_file);
		}else { 
			type = get_type_str(lids[i].type);
			printf("\t%10s\t%8s:%d\t%10s\n",s_file,lids_type_desc[type],lids[i].inherit,o_file);
		}

		if( type == LIDS_EXTEND) {
			printf("\t(%d,%d)\t%d\t(%d,%d)\n",lids[i].s_ino,lids[i].s_dev,lids[i].type,lids[i].o_ino,lids[i].o_dev);
		}
	}	
	printf("\n\n");
}
/*
 *	lids_search_file()
 *
 *	lids search the file by given file and type.
 *
 */
int lids_search_file(char *s_file,char *o_file) 
{
	int i ,flag,number=0;

	for ( i = 0; i<last ; i++ ) {
		flag = 0;
		if (*s_file != '\0') {
		 	flag = strcmp(lids[i].s_file , s_file);
		}
		if (*o_file != '\0')
		 	flag |= strcmp(lids[i].o_file , o_file);
		if(!flag) {
			number++;
			lids[i].type = LIDS_DELETE_FLAG;
		}
	}
        printf("delete %d items\n",number);
	return number;
}

/*
 *	lids_search_inode()
 *
 *	lids search the file by given inode , dev  and type.
 *
 */
int lids_search_inode(struct stat s_st,struct stat o_st) 
{
	int i;

	for ( i = 0; i<last ; i++ ) {
		if ( lids[i].s_ino == s_st.st_ino && 
		     lids[i].s_dev == s_st.st_dev &&
		     lids[i].o_ino == o_st.st_ino && 
		     lids[i].o_dev == o_st.st_dev 
		     ) {   
			return lids[i].type;
		}
	}	
	return LIDS_DELETE_FLAG;
}
void lids_del_file(char *subject_file,char *object_file)
{
	int i;
	char canonical_s_path[PATH_MAX];
	char canonical_o_path[PATH_MAX];

	lids_read_conf();

	i = lids_search_file(subject_file,object_file);
		
	if(!i) {
		realpath(subject_file,canonical_s_path);
		realpath(object_file,canonical_o_path);
		i = lids_search_file(canonical_s_path,canonical_o_path);
		if (!i) 
			exit_error(2,"the file does not exit in "LIDS_CONF_FILE);
	}

	lids_write_file();
}
/*
 *	get special type from typestr, when lidsadm -A -s ..-o .. -t -j ..
 */ 
int get_cap_number(char *typestr)
{
	int i;
	for(i=0;thecaps[2*i]!=NULL;i++) {
		if(strcmp(typestr,thecaps[2*i]) == 0 )
				return i;
	}	
	return -1;
}
/*
 *
 */
int get_object_type(char *filename)
{
	int i;
	char o_file[1024];
	char *p=NULL;
		
	strcpy(o_file,filename);
	strcat(o_file,"/");

	while( p != o_file) { 
		p = rindex(o_file,'/');
		if(!p)
			break;
		*p = '\0';
		for(i=0; i<last;i++) {
			if(*(lids[i].s_file)=='\0' && strcmp(o_file,lids[i].o_file) == 0 ) {
				return lids[i].type;

			}
		}
	}
	return -1;
	
}
/*
 *	lids_add_file 
 */ 
void lids_add_file(char *subject_file, char *object_file,int type,int inherit,int domain)
{
	struct stat s_st,o_st,z_st;
	char canonical_s_path[PATH_MAX];
	char canonical_o_path[PATH_MAX];
	int  default_rule=-1;
	int  sys_cap = 0;	/* 1 for system cap, 0 for individual cap */

	lids_read_conf();

        if ( last >= 1024 ) {
                exit_error(2, "cannot exceed 1024 entries.");
        }
	
#ifdef DEBUG	
	printf("sub=[%s],type=%d\n",subject_file,type);
#endif
	if(*subject_file == '\0' ) {
		sys_cap = 1;
	}

	if(!sys_cap) {	
		realpath(subject_file,canonical_s_path);
		if ( stat(canonical_s_path,&s_st) ) {
			exit_error(2, " cannot find the subject file");
		}
	}
	else {
		s_st.st_ino = s_st.st_dev = 0;
	}
	if(type != LIDS_CAP) {
		realpath(object_file,canonical_o_path);
		if(stat(canonical_o_path,&o_st) ) {
			exit_error(2, " cannot find the object file");
		}
	}
	else {
		o_st.st_ino = -1;	/* special type's ino */
		if((int)(o_st.st_dev = get_cap_number(object_file)) < 0 ) {
			exit_error(2, "special type must be one of CAP_XXX_XXX. use lidsadm -h for details.");
		}
		strcpy(canonical_o_path,object_file);
	}
	/* check the subject */
	if(!sys_cap) {
		if((default_rule = get_object_type(canonical_s_path)) == -1)
			exit_error(2,"the subject file is not protected");
		if(default_rule > LIDS_READ )
			exit_error(2,"the subject files must be protected by READ or DENY"); 
	}
	/* check the object */
	if(type != LIDS_CAP) {
		if( (default_rule = get_object_type(canonical_o_path)) == -1) {
			if(!sys_cap)
				exit_error(2,"you must define the default rules for object files");
		}
	}
	
	if(domain) type = -type;

	if (!(lids_search_inode(s_st,o_st) != type ||
		lids_search_inode(s_st,o_st) == LIDS_DELETE_FLAG))
			exit_error(2,"a file with the same inode and dev is already referenced.");

	lids[last].s_ino= s_st.st_ino;
	lids[last].s_dev = s_st.st_dev;
	lids[last].o_ino= o_st.st_ino;
	lids[last].o_dev = o_st.st_dev;
#ifdef DEBUG	
	printf("canonical_o_file=%s,o_ino =%d\n",canonical_o_path,o_st.st_ino);
#endif
	lids[last].type = type;
	lids[last].inherit = inherit;

	strcpy(lids[last].s_file,canonical_s_path);
	strcpy(lids[last].o_file,canonical_o_path);
		
	last++;
	lids_write_file();
}

void lids_update()
{
	struct stat s_st,o_st;
	int i;
	lids_read_conf();
	

	for (i=0; i<last; i++) {
		/*
		 *	check subject files
		 */ 
		if (*lids[i].s_file != '\0' ) {
			if(stat(lids[i].s_file,&s_st)) {
				printf("%s doesn't exist anymore. not removed.\n",lids[i].s_file);
			}
			else {
				if ((s_st.st_ino != lids[i].s_ino) || (s_st.st_dev != lids[i].s_dev) ) {
				printf("subject file %s was (%u:%u inode %lu) instead of (%u:%u %lu). corrected.\n",
				       lids[i].s_file,
				       MAJOR(lids[i].s_dev),MINOR(lids[i].s_dev),lids[i].s_ino,
				       MAJOR((unsigned long)s_st.st_dev),MINOR((unsigned long)s_st.st_dev),s_st.st_ino);
				lids[i].s_ino = s_st.st_ino;
				lids[i].s_dev = s_st.st_dev;
				}
			}	
		}
		/*
		 *	check object files
		 */ 
		if (*lids[i].o_file != '\0' && stat(lids[i].o_file,&o_st)) {
			if(get_cap_number(lids[i].o_file) < 0 && lids[i].o_ino == -1) {
			printf("%s doesn't exist anymore. not removed.\n",lids[i].o_file);
			}
		}
		else {
			if (o_st.st_ino != lids[i].o_ino || o_st.st_dev != lids[i].o_dev) {
				printf("object file %s was (%u:%u inode %lu) instead of (%u:%u %lu). corrected.\n",
				       lids[i].o_file,
				       MAJOR(lids[i].o_dev),MINOR(lids[i].o_dev),lids[i].o_ino,
				       MAJOR((unsigned long)o_st.st_dev),MINOR((unsigned long)o_st.st_dev),o_st.st_ino);

				lids[i].o_ino = o_st.st_ino;
				lids[i].o_dev = o_st.st_dev;
			}
		}
	}
	lids_write_file();
}

void
exit_error (int status, char *msg)
{
	fprintf (stderr, "%s: %s\n", program, msg);
	if(status == 3 )
		perror("reason:");
	printf("\n");
	exit (status);
}
void
exit_version ()
{
       printf ("%s v0.10 for LIDS project\n"
               "Use '%s -h' for help\n"
               ,program,program);
       exit(1);
}

void
exit_normal ()
{
       printf ("%s for LIDS project\n"
               "Use '%s -h' for help or '%s -v' for version\n"
               ,program,program,program);
       exit(1);
}

void
exit_help ()
{
	int i;
	printf ("%s for LIDS project\n" 
		"       Huagang Xie<xie@gnuchina.org>\n"
		"       Philippe Biondi <philippe.biondi@webmotion.net>\n\n"
		"Usage: %s -A [-s subject] -o object [-d] -j ACTION\n"
		"       %s -D [-s file] [-o file] \n"
		"       %s -Z\n"
		"       %s -U\n"
		"       %s -L\n"
		"       %s -P\n"
		"       %s -[S|I] -- [+|-][LIDS_FLAG] [...]\n"
		"       %s -v\n"
#ifdef VIEW
		"       %s -V\n"
#endif
		"       %s -h\n\n"
		"Commands:\n"
		"       -A  To add an entry\n"
		"       -D  To delete an entry\n"
		"       -Z  To delete all entries \n"
		"       -U  To update dev/inode numbers\n"
		"       -L  To list all entries \n"
		"       -P  To encrypt a password with RipeMD-160\n"
		"       -S  To submit a password to switch some protections\n"
		"       -I  To switch some protections without submitting password (sealing time)\n"
		"       -v  To show the version\n"
#ifdef VIEW
		"       -V  To view current LIDS state (caps/flags)\n"
#endif
		"       -h  To list this help \n\n"
		"subject:\n"
		"       can be any program, must be a file\n" 
		"object:\n"
		"       can be a file, directory or special device (e.g. MEM, HD, NET, IO,\n"
                "                                                        HIDDEN, KILL)\n"
		"ACTION:\n"
		"       DENY    deny access\n"
		"       READ    read only\n" 
		"       APPEND  append only\n"
	        "       WRITE   writable\n"
		"       GRANT   grant capability to subject\n"
		"       IGNORE  ignore any permissions set on this object\n"
		
		"TYPE:\n"	
		"       -d      the object is an EXEC Domain\n"
#ifdef VIEW
		,program
#endif
		,program,program,program,program,program,program,program,program,program,program);

	printf("\nAvailable capabilities:\n");
	for (i=0; thecaps[2*i] != NULL; i++) 
		printf("%20s %s\n",thecaps[2*i],thecaps[2*i+1]);
	printf("\nAvailable flags:\n");
	for (i=0; theflags[2*i] != NULL; i++) 
		printf("%20s %s\n",theflags[2*i],theflags[2*i+1]);

	exit(1);
}
void lids_make_rmd160_passwd()
{
	char passwd[64*2];
	int times=0;
	int ok;
	FILE *fd;
	
	memset(passwd,'\0',64*2);
	while(times < 3) {
		times++;	
		if ((ok=read_rmd160_passwd(passwd,1,2))==0)  
			break;
	}
	
	if (ok==0) {
		fd = fopen(LIDS_PW_FILE,"w");
		if (fd == NULL) return;	
		fwrite(passwd,1,strlen(passwd),fd);
		fclose(fd);
		printf("wrote password to %s\n", LIDS_PW_FILE); 
	}
	
}
	

void lids_set_caps(int optind, int argc, char *argv[],lids_locks_t *locks)

{
	int i;
	
	for (i=optind; i < argc; i++) {
		int capnum=-1;
		char *capstr=NULL, *capdescstr=NULL;
		int flagnum=-1;
		char *flagstr=NULL, *flagdescstr=NULL;
		int j=0;
			
		while (thecaps[j] != NULL) {
			if (!strcmp(argv[i]+1, thecaps[j])) {
				capnum = j/2;
				capstr = thecaps[j];
				capdescstr = thecaps[j+1];
				break;
			}
			j += 2;
		}
		
		if (capnum == -1) {
			j=0;
			
			while (theflags[j] != NULL) {
				if (!strcmp(argv[i]+1, theflags[j])) {
					flagnum = j/2;
					flagstr = theflags[j];
					flagdescstr = theflags[j+1];
					break;
				}
				j += 2;
			}
		}
			
		if ( ((capnum == -1) || (capstr == NULL)) &&
		     ((flagnum == -1) || (flagstr == NULL)) ) {
			fprintf(stderr, "    %s: invalid capability/flag\n", argv[i]);
			exit(1);
		}

		switch (argv[i][0]) {
		case '+':
			if (flagnum == -1)
				cap_raise(locks->cap_bset, capnum);
			else
				flag_raise(locks->flags, flagnum);
			break;
		case '-':
			if (flagnum == -1)
				cap_lower(locks->cap_bset, capnum);
			else
				flag_lower(locks->flags, flagnum);
			break;
		default:
			fprintf(stderr, "%s: invalid option\n", argv[i]);
			exit(1);
		}
			
	}
}

void lids_switch(int optind, int argc, char *argv[])
{
	int lk ;
	char passwd[BUFSIZ];
	lids_locks_t locks,locks2;

	if ((lk=open(LIDS_LOCKS,O_RDWR)) == -1) {
		perror("open");
		exit_error (2, "cannot open " LIDS_LOCKS);
	}
	if (read(lk,&locks,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "cannot read " LIDS_LOCKS);
	}

	lids_set_caps(optind,argc,argv,&locks);

	locks.magic1=LIDS_MAGIC_1;
	locks.magic2=LIDS_MAGIC_2;
	locks.magic3=LIDS_MAGIC_3;
	locks.magic4=LIDS_MAGIC_4;

	read_rmd160_passwd(passwd,0,1);  

	strncpy(locks.passwd,passwd,64);

	if (write(lk,&locks,sizeof(lids_locks_t))==-1) {
		perror("write");
		exit_error (2, "cannot write " LIDS_LOCKS);
	}

	if (read(lk,&locks2,sizeof(lids_locks_t))==-1) {
		perror("reread");
		exit_error (2, "cannot reread " LIDS_LOCKS);
	}

	close(lk);

	/* 
	 * Dont test RELOAD_CONF as 
	 * it is always read as 0
	 */
	locks.flags &= ~4;

	if ((locks.cap_bset != locks2.cap_bset) || 
	    (locks.flags != locks2.flags)) { 
		fprintf(stderr,"switching LIDS failed\n");
	}

}

void lids_init(int optind, int argc, char *argv[])
{
	int fd ;
	lids_locks_t locks,locks2;
	locks.cap_bset=0;
	locks.flags=0;
	
	if ((fd=open(LIDS_LOCKS,O_RDWR)) == -1) {
		perror("open");
		exit_error (2, "cannot open " LIDS_LOCKS);
	}
	if (read(fd,&locks,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "cannot read " LIDS_LOCKS);
	}

	lids_set_caps(optind,argc,argv,&locks);
	
	locks.magic1=LIDS_MAGIC_1;
	locks.magic2=LIDS_MAGIC_2;
	locks.magic3=LIDS_MAGIC_3;
	locks.magic4=LIDS_MAGIC_4;
	locks.passwd[0]=0;

	if (write(fd,&locks,sizeof(lids_locks_t))==-1) {
		perror("write");
		exit_error (2, "cannot write " LIDS_LOCKS);
	}

	if (read(fd,&locks2,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "cannot reread " LIDS_LOCKS);
	}
	close(fd);

	/* 
	 * Dont test RELOAD_CONF test as 
	 * it is always read as 0
	 */
	locks.flags &= ~4;

/*	if ((locks.cap_bset != locks2.cap_bset) || */
	  if  ((locks.flags != locks2.flags)) 
		fprintf(stderr,"LIDS init failed\n");
}
	
#ifdef VIEW
void lids_view()
{
	int fd ;
	lids_locks_t locks;
	int i;
	locks.cap_bset=0;
	locks.flags=0;

	if ((fd=open(LIDS_LOCKS,O_RDWR)) == -1) {
		perror("open");
		exit_error (2, "cannot open " LIDS_LOCKS);
	}
	if (read(fd,&locks,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "cannot read " LIDS_LOCKS);
	}
	close(fd);

	for(i=0;(i < 8*sizeof(kernel_cap_t)) && (thecaps[2*i] != NULL);i++)
		    printf("%30s %i\n",thecaps[2*i],(locks.cap_bset >> i) & 1);
	for(i=0;(i < 8*sizeof(lids_flags_t)) && (theflags[2*i] != NULL);i++) 
		    printf("%30s %i\n",theflags[2*i],(locks.flags >> i) & 1);

}
#endif
